module.exports = {
    ...require('typescript'),
    transpileModule() {
      return {
        outputText: 'emit from shared-config custom compiler',
        sourceMapText: '{}',
      };
    },
    createProgram() {
      throw 'emit from shared-config custom compiler';
    }
  };
