module.exports.create = function () {
    return {
      transpile() {
        return {
          outputText: 'emit from root custom transpiler',
          sourceMapText: '{}',
        };
      },
    };
  };
