// These files are resolved by the typechecker
import * as tsmii from './typescript-module-imported-internally'
export {tsmii}
import * as jsmii from './javascript-module-imported-internally'
export {jsmii}
// These files are unknown to the compiler until required.
export const tsmri = require('./typescript-module-required-internally')
export const jsmri = require('./javascript-module-required-internally')
