declare module 'lib_greeter' {
  /** API for saying hello to everyone. */
  export class Greeter {
    /** Returns a friendly salutation and a remark about the weather */
    sayHello(): string;
  }
}
